//---------------------------------------------------------------------------
#ifndef AssemblerH
#define AssemblerH
//---------------------------------------------------------------------------

#include "Instruction.h"
#include "SymTable.h"
#include "Preprocessor.h"

enum AsmError
{
  PPOPEN = 100, PPNOLINES, PPNOFILES, PPUNREQDIR, PPFNAMEEXP,
    PPFILEINC, PPIDENEXP, PPDEFFULL,
  PLINELEN = 200, PSYMLEN, PSYMREQ, PSYMCHAR, PSYMNUM, PDUPSYM,
    PSYMFULL, POBUFOVER, PESYMCHAR, PEXTINAR, PMULTIEXT, PEXTDD,
    PEXTBCONST, PRELJUMP, PGENRELOC, PEXTNOJR, PSYMNDEF,
    PRELBCONST, PNOPADBACK, PNOOPENBIN, PLLTOOLONG, PNOLTOP,
    PDDNOTCONST, PPADNOTCONST,
  PILLEXPR = 300, PDIVZERO, PNOOPENP, PNOCLOSEP, PCHARCONST,
    PUNREQDIR, PEXPREXP, PINVESC, PNOCQUOTE,
  ITOPEN = 400, ITPARSE, ITOPSIZE, ITLINELEN, ITUNRECINSTR,
    ITUNRECARG, ITMAXINSTR, ITNOEXRELOC
};

enum AsmWarning
{
  WPDUPSYM = 200, WPSYMTRUNC,
  WPTRUNCW = 300, WPTRUNCB,
  WITTRUNCW = 400, WITTRUNCB, WITILBITVAL
};

class CAssembler
{
  public:
    CAssembler() {}
    ~CAssembler() {}
    void Error(enum AsmError n, const char *s);
    void ErrorI(enum AsmError n, int i);
    void Warning(enum AsmWarning n, const char *s);
    void Message(const char *m, const char *s);
    void NextLine();
    char *Filename();
    int  LineNumber();
    int  PCounter();
    bool Errors();
    char *LastLocal();
    bool Assemble(const char *tfile, const char *fname,
                  const char *oname, bool werror);
    int  Parse(const char *c, int *e, int o);
    bool External(const char *c, bool obj);
    bool Local(const char *c, bool obj);
    bool ExtRef(const char *c, bool obj);

  private:
    char *GetExtRef(const char *c, char *e);
    int  Directive(char *c, bool addsym, bool output);

    CInstruction _instruction;
    CSymTable _symtable;
    CPreprocessor _preprocessor;

    char _filename[256],
         _lastlocal[MAXSYM];
    int _linenumber,
        _pcounter,
        _errors,
        _warnings;
};

extern CAssembler assembler;

#endif
