//---------------------------------------------------------------------------
#include <string.h>
#include <ctype.h>
#pragma hdrstop

#include "Instruction.h"
#include "Assembler.h"
#include "ObjectFile.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

CInstruction::CInstruction()
{
  _instructions = 0;
}

void CInstruction::LoadTable(const char *fname)
{
  FILE *fp;
  int x;
  char *p;
  char line[MAXLINE],
       opcode[MAXLINE];
  _Instr i;

  assembler.Message("Loading instruction table \"%s\"", fname);
  fp = fopen(fname, "r");
  if (fp == NULL)
  {
    assembler.Error(ITOPEN, fname);
    return;
  }
  while (!feof(fp))
  {
    fgets(line, sizeof(line), fp);
    if ((feof(fp)) && (strlen(line) == 0))
      break;
    if (strlen(line) == sizeof(line))
    {
      assembler.ErrorI(ITLINELEN, sizeof(line) - 1);
      return;
    }
    if ((p = strchr(line, '\n')) != NULL)
      *p = '\0';
    if ((p = strchr(line, ';')) != NULL)
      *p = '\0';
    if (strlen(line) > 0)
    {
      if (sscanf(line, "%s %s %s %i %i", &i.instr, &i.args,
                  &opcode, &i.bytes, &i.time) != 5)
        assembler.Error(ITPARSE, "");
      else
      {
        if ((i.bytes < 1) || (i.bytes > MAXOPSIZE))
          assembler.ErrorI(ITOPSIZE, MAXOPSIZE);
        else
        {
          x = 0;
          while (sscanf(opcode + (x * 2), "%2X", &i.opcode[x]) == 1)
            x++;
          if (i.args[0] == '\"')
            i.args[0] = '\0';
          i.opsize = x;
          if (_instructions + 1 == INSTRMAX)
          {
            assembler.ErrorI(ITMAXINSTR, INSTRMAX);
            return;
          }
          _instr[_instructions++] = i;
        }
      }
    }
    assembler.NextLine();
  }
  fclose(fp);
}

int CInstruction::GetSize(const char *c, char *e)
{
  char b[MAXLINE];
  int n;
 
  n = FindInstr(c, b);
  if (e != NULL)
    strcpy(e, b);
  if (n == -1)
    return 0;

  return _instr[n].bytes;
}

int CInstruction::WriteOpcode(char *o)
{
  char e[MAXLINE];
  unsigned char c;
  int n, i, r;
  bool x;

  n = FindInstr(o, e);
  x = false;
  if (n == -1)
    return 0;
  objectfile.OutputInstr();
  if (strchr(_instr[n].args, '!') == NULL)
    for(i = _instr[n].opsize - 1; i >= 0; i--)
      objectfile.OutputByte(_instr[n].opcode[i]);
  else
  {
    r = assembler.Parse(e, NULL, 0);
    for(i = _instr[n].opsize - 1; i > 0; i--)
      objectfile.OutputByte(_instr[n].opcode[i]);
    if (r >= 8)
      assembler.Warning(WITILBITVAL, "");
    objectfile.OutputByte((char)(_instr[n].opcode[0] + ((r % 8) << 3)));
  }
  i = _instr[n].bytes - _instr[n].opsize;
  if (i > 0)
  {
    r = assembler.Parse(e, NULL, 0);
    if (strchr(_instr[n].args, '@') != NULL)
    {
      if (assembler.External(e, false))
        assembler.Error(PEXTNOJR, "");
      r -= assembler.PCounter() + _instr[n].bytes;
      if ((signed char)r != r)
        assembler.Error(PRELJUMP, "");
      objectfile.OutputByte((signed char)r);
    }
    else
    {
      if (i == 1)
      {
        if ((assembler.External(e, false)) || (assembler.ExtRef(e, false)))
          assembler.Error(PEXTBCONST, "");
        if (assembler.Local(e, false))
          assembler.Error(PRELBCONST, "");
        if (r > 255)
        assembler.Warning(WITTRUNCB, "");
        objectfile.OutputByte((char)(r % 256));
      }
      else
      {
        if (r > 65535)
          assembler.Warning(WITTRUNCW, "");
        objectfile.OutputByte((char)(r % 256));
        objectfile.OutputByte((char)(r >> 8));
        x = assembler.External(e, true);
        assembler.ExtRef(e, true);
        if (x)
        {
          c = _instr[n].opcode[_instr[n].opsize - 1];
          if ((c != 0x21) && (c != 0xC3) && (c != 0xCD))
            assembler.Error(ITNOEXRELOC, "");         
          objectfile.OutputByte(0);
        }
        assembler.Local(e, true);
      }
    }
  }

  return _instr[n].bytes + x;
}

int CInstruction::FindInstr(const char *c, char *e)
{
  char instr[MAXLINE];
  unsigned int s, i, a, p;
  int n;
  bool matched;

  for(s = 0; c[s] == ' '; s++) {}
  if (c[s] == '\0')
    return -1;

  for(i = s; (c[i] != ' ') && (i < strlen(c)); i++)
    instr[i - s] = c[i];
  instr[i - s] = '\0';

  for(s = i; c[s] == ' '; s++) {}

  if (e != NULL)
    e[0] = '\0';
  matched = false;
  for(n = 0; n < _instructions; n++)
    if (!stricmp(_instr[n].instr, instr))
    {
      i = s;
      for(a = 0; a < strlen(_instr[n].args); a++)
      {
        if ((_instr[n].args[a] != '*') &&
            (_instr[n].args[a] != '!') &&
            (_instr[n].args[a] != '@'))
        {
          if (_instr[n].args[a] != toupper(c[i++]))
            break;
        }
        else
        {
          if (a + 1 == strlen(_instr[n].args))
          {
            if (e != NULL)
              strcpy(e, &c[i]);
            return n;
          }
          p = i;
          if (_instr[n].args[a] == ')')
            a++;
          while ((c[i] != _instr[n].args[a + 1]) && (i < strlen(c)))
            i++;
          if (e != NULL)
          {
            memcpy(e, &c[p], i - p);
            e[i - p] = '\0';
          }
        }
      }
      while (c[i] == ' ')
        i++;
      if ((a == strlen(_instr[n].args)) && (i == strlen(c)))
        return n;
      else
          matched = true;
    }

  if (matched)
    assembler.Error(ITUNRECARG, instr);
  else
      assembler.Error(ITUNRECINSTR, instr);

  return -1;
}
