//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#pragma hdrstop

#include "ObjectFile.h"
#include "SymTable.h"

CObjectFile objectfile;

void CObjectFile::New()
{
  _reloc = 1;
  _outputsize = 0;
  _line = 0;
  _table[0].addr = 0;
}

void CObjectFile::OutputByte(const char c)
{
  if (_outputsize == MAXOUTPUT)
  {
    assembler.ErrorI(POBUFOVER, MAXOUTPUT);
    _outputsize = 0;
  }
  _output[_outputsize++] = c;
}

void CObjectFile::OutputInstr()
{
  _linenum[_line].addr = assembler.PCounter();
  _linenum[_line++].line = assembler.LineNumber();
}

void CObjectFile::AddLocal()
{
  _table[0].address[_table[0].addr++] = assembler.PCounter() + 1;
}
void CObjectFile::AddLocalAddr(int addr)
{
  _table[0].address[_table[0].addr++] = addr;
}

void CObjectFile::AddExt(const char *c)
{
  int i;

  for(i = 1; i < _reloc; i++)
    if (!strcmp(c, _table[i].name))
    {
      _table[i].address[_table[i].addr++] = assembler.PCounter() + 1;
      return;
    }
  _table[_reloc].addr = 1;
  _table[_reloc].address[0] = assembler.PCounter() + 1;
  strcpy(_table[_reloc++].name, c);
}

bool CObjectFile::Write(const char *fname, CSymTable &sym)
{
  FILE *fp;
  int i, x;
  char s[MAXSYM];

  fp = fopen(fname, "wb");
  if (fp == NULL)
  {
    assembler.Error(PPOPEN, fname);
    return false;
  }
  fputw(_outputsize, fp);
  fwrite(_output, _outputsize, 1, fp);
  fputw(_table[0].addr, fp);
  for(x = 0; x < _table[0].addr; x++)
    fputw(_table[0].address[x], fp);
  fputw(_reloc - 1, fp);
  for(i = 1; i < _reloc; i++)
  {
    fputc(strlen(_table[i].name), fp);
    fwrite(_table[i].name, strlen(_table[i].name) + 1, 1, fp);
    fputw(_table[i].addr, fp);
    for(x = 0; x < _table[i].addr; x++)
      fputw(_table[i].address[x], fp);
  }
  fputw(sym.FirstLocal(), fp);
  while (sym.NextLocal(s, i))
  {
    fputc(strlen(s), fp);
    fwrite(s, strlen(s) + 1, 1, fp);
    fputw(i, fp);
  }
  fputw(_line, fp);
  for(i = 0; i < _line; i++)
  {
    fputw(_linenum[i].line, fp);
    fputw(_linenum[i].addr, fp);
  }
  fclose(fp);
//  Dump();

  return true;
}

void CObjectFile::Dump()
{
  int i, x;
  char line[256];

  sprintf(line, "Output size: %i", _outputsize);
  assembler.Message(line, "");
  assembler.Message("Local relocation table:", "");
  for(i = 0; i < _table[0].addr; i++)
  {
    sprintf(line, "  Addr: %04x", _table[0].address[i]);
    assembler.Message(line, "");
  }
  assembler.Message("External relocation table:", "");
  for(i = 1; i < _reloc; i++)
  {
    assembler.Message("  Var: %s", _table[i].name);
    for(x = 0; x < _table[i].addr; x++)
    {
      sprintf(line, "    Addr: %04x", _table[i].address[x]);
      assembler.Message(line, "");
    }
  }
/*
  assembler.Message("Line numbers:", "");
  for(i = 0; i < _line; i++)
  {
    sprintf(line, "  Line: %5i    Addr: %04x",
            _linenum[i].line, _linenum[i].addr);
    assembler.Message(line, "");
  }
*/
}

void CObjectFile::fputw(int c, FILE *fp)
{
  fputc(c % 256, fp);
  fputc(c >> 8, fp);
}

//---------------------------------------------------------------------------
#pragma package(smart_init)
