//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#pragma hdrstop

#include "SymTable.h"
#include "Assembler.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)


CSymTable::CSymTable()
{
  _table = new struct _SymEntry[MAXSYMENT];
  _entries = 0;
}

void CSymTable::Add(const char *n, int v, bool e,
                    bool l, bool p, const char *r)
{
  char t[MAXSYM];

  if (*n == '@')
  {
    if (strlen(n) + strlen(assembler.LastLocal()) + 1 >= MAXSYM)
    {
      assembler.Error(PLLTOOLONG, "");
      return;
    }
    if (!strlen(assembler.LastLocal()))
    {
      assembler.Error(PNOLTOP, "");
      return;
    }
    strcpy(t, "@");
    strcat(t, assembler.LastLocal());
    strcat(t, n);
    p = false;
  }
  else
    strcpy(t, n);
  if (v > 65535)
    assembler.Warning(WPSYMTRUNC, "");
  v %= 65536;
  if (Exists(t))
  {
    if (Get(t) != v)
      assembler.Error(PDUPSYM, n);
    else
      assembler.Warning(WPDUPSYM, n);
    return;
  }

  if (_entries == MAXSYMENT)
  {
    assembler.ErrorI(PSYMFULL, MAXSYMENT);
    return;
  }
  strcpy(_table[_entries].name, t);
  _table[_entries].value = v;
  _table[_entries].line = assembler.LineNumber();
  _table[_entries].external = e;
  _table[_entries].local = l;
  _table[_entries].exp = p;
  if (r != NULL)
    _table[_entries].extref = Find(r);
  else
    _table[_entries].extref = -1;
  _entries++;
}

int CSymTable::Get(const char *n)
{
  char t[MAXSYM];
  int i;

  if (*n == '@')
  {
    if (strlen(n) + strlen(assembler.LastLocal()) + 1 >= MAXSYM)
    {
      assembler.Error(PLLTOOLONG, "");
      return 0;
    }
    if (!strlen(assembler.LastLocal()))
    {
      assembler.Error(PNOLTOP, "");
      return 0;
    }
    strcpy(t, "@");
    strcat(t, assembler.LastLocal());
    strcat(t, n);
  }
  else
    strcpy(t, n);
  i = Find(t);
  if (i != -1)
    return _table[i].value;
  assembler.Error(PSYMNDEF, t);

  return 0;
}

char *CSymTable::GetExt(const char *n)
{
  int i;

  i = Find(n);
  if (i == -1)
    return NULL;
  i = _table[i].extref;
  if (i != -1)
    return _table[i].name;

  return NULL;
}

bool CSymTable::Exists(const char *n)
{
  if (Find(n) != -1)
    return true;

  return false;
}

void CSymTable::Dump()
{
  char line[256],
       buf[256];
  int  i;

  assembler.Message("Symbol table:", "");
  for(i = 0; i < _entries; i++)
  {
    if (_table[i].external)
      sprintf(line, "  %s = *", _table[i].name);
    else
      sprintf(line, "  %s = %i", _table[i].name, _table[i].value);
    if (_table[i].local)
      strcat(line, " @");
    if (_table[i].extref != -1)
    {
      sprintf(buf, " (%s)", _table[_table[i].extref].name);
      strcat(line, buf);      
    }
    assembler.Message(line, "");
  }
}

bool CSymTable::External(const char *n)
{
  int i;

  i = Find(n);
  if ((i == -1) || (!_table[i].external))
    return false;

  return true;
}

bool CSymTable::Local(const char *n)
{
  char t[MAXSYM];
  int i;

  if (*n == '@')
  {
    if (strlen(n) + strlen(assembler.LastLocal()) + 1 >= MAXSYM)
    {
      assembler.Error(PLLTOOLONG, "");
      return false;
    }
    if (!strlen(assembler.LastLocal()))
    {
      assembler.Error(PNOLTOP, "");
      return false;
    }
    strcpy(t, "@");
    strcat(t, assembler.LastLocal());
    strcat(t, n);
  }
  else
    strcpy(t, n);
  i = Find(t);
  if ((i == -1) || (!_table[i].local))
    return false;

  return true;
}

int CSymTable::FirstLocal()
{
  int i, n;

  n = 0;
  for(i = 0; i < _entries; i++)
//    if (_table[i].exp) && (_table[i].name[0] != '@'))
    if (_table[i].exp)
      n++;
  _currlocal = 0;
//  Dump();

  return n;
}

bool CSymTable::NextLocal(char *n, int &v)
{
  while (_currlocal < _entries)
  {
    if (_table[_currlocal].exp)
    {
      strcpy(n, _table[_currlocal].name);
      v = _table[_currlocal].value;
      _currlocal++;
      return true;
    }
    _currlocal++;
  }

  return false;
}

int CSymTable::Find(const char *n)
{
  int i;

  for(i = 0; i < _entries; i++)
    if (!strcmp(_table[i].name, n))
      return i;

  return -1;
}

void CSymTable::BeginGet()
{
  _currget = 0;
}

bool CSymTable::NextGet(char *n, int &v)
{
  while (_currget < _entries)
  {
    if (!_table[_currget].external)
    {
      strcpy(n, _table[_currget].name);
      v = _table[_currget].value;
      _currget++;
      return true;
    }
    _currget++;
  }

  return false;
}

