; Base code for CGB
; (c) 1999 ACZ

;rst equates
_ldhlind			=		$00			;ld hl,(hl)
_addhla				=		$08			;add hl,a
_cphlde				=		$10			;cp hl,de
_subhlde			=		$18			;sub hl,de (not sbc hl,de)
_exdehl				=		$20			;ex de,hl
_cpagejp			= 		$28			;cross-page jp
_cpagecall			=		$30			;cross-page call
_cpageldhl			= 		$38			;cross-page ld hl,*

_main				=		$150		;program start

;ram equates
_video				=		$8000		;video memory
_obpa				=		$8000		;object pattern area
_bgpa				=		$8800		;background pattern area
_wnpa				=		$9000		;window & background pattern area
_bg1				=		$9800		;primary background tile map area
_bg2				=		$9c00		;second background tile map area
_win1				=		$9c00		;primary window tile map area
_win2				=		$9800		;second window tile map area
_eew				=		$a000		;external expanded wram
_bank0				=		$c000		;static bank 0
_bankv				=		$d000		;swappable banks 0-7

;$e000 - $fdff is reserved
_oam				=		$fe00		;object attribute memory
_io					=		$ff00		;i/o
_vblank				=		$ff80		;vblank interrupt handler

;note that the stack works down from $fffe


;port equates (one byte)
p1					=		$00		;keypad
sb					=		$01		;serial data
sc					=		$02		;serial i/o control
div					=		$04		;divide register
tima				=		$05		;timer counter
tma					=		$06		;timer modulo
tac					=		$07		;timer control
if					=		$0f		;interrupt flag

lcdc				=		$40		;lcd control
stat				=		$41		;lcd status
scy					=		$42		;scroll y
scx					=		$43		;scroll x
ly					=		$44		;lcd y
lyc					=		$45		;ly compare
dma					=		$46		;dma start address
bgp					=		$47		;background palette data
obp0				=		$48		;object palette 0 data
obp1				=		$49		;object palette 1 data
wy					=		$4a		;window y position
wx					=		$4b		;window x position
ie					=		$ff		;interrupt enable

p0					=		$30
p14					=		$20
p15					=		$10

;color only ports
key1				=		$4d		;cpu speed (1.05 or 2.1 MHz)
									;set bit 0 - enable switch
									;read bit 7 -current speed
									;execute STOP for switch to
									;activate (all IE disabled)
									;speed 0 selected at reset
vbk					=		$4f		;video bank specification
									;bit 0 - reset bank 0
									;bit 1 - set bank 1
									;bank 0 selected at reset
hdma1				=		$51		;hblank dma transfer source
hdma2				=		$52		;bits 4-7 - low source
									;hdma1 and hdm2 combine to
									;form 12 msb of addr between
									;$0000-$7fff or $a000-$dfff
hdma3				=		$53		;hblank dam transfer dest
									;bits 0-5 - high dest
hdma4				=		$54		;bits 4-7 - low dest
									;hdma3 and hdma4 combine to
									;form 9 msb of addr between
									;$8000 and $9fff
hdma5				=		$55		;hblank dma transf quantity
									;bits 0-6 - # of lines + 1
									;copies 16(n+1) bytes
									;bit 7 - transfer start
									;1 - start hblank dma
									;0 - end hblank dma
rp					=		$56		;infra red port
bcps				=		$68		;bg write specification
									;bit 0 - 0 low / 1 high
									;bits 1-2 - palette data #
									;bits 3-5 - palette #
									;bit 7 - 0 fixed / 1 cycle
bcpd				=		$69		;background write data
ocps				=		$6b		;obj write specification
									;bits same as bcps
ocpd				=		$6c		;object write data
svbk				=		$70		;work memory bank
									;0-1 bank 1
									;2-7 bank 2-7

.pad _ldhlind			;ld hl,(hl)
 push af
 ld a,(hl)
 inc hl
 ld l,(hl)
 ld h,a
 pop af
 ret

.pad _addhla			;add hl,a
 add a,l
 ld l,a
 ret nc
 inc h
 ret

.pad _cphlde			;cp hl,de
 ld a,h
 xor d
 ret nz
 ld a,l
 xor e
 ret

.pad _subhlde			;sub hl,de
 ld a,l
 sub e
 ld a,h
 sbc a,d
 ret

.pad _exdehl			;ex de,hl
 ld a,l
 sub e
 ld a,h
 sbc a,d
 ret

.pad _cpagejp			;cross-page jp
 call _cpagerst
 jr _cpagejp_

.pad _cpagecall			;cross-page call
 call _cpagerst
 jr _cpagecall_

.pad _cpageldhl			;cross-page ld hl,*
 call _cpagerst
 jr _cpageldhl_ 

.pad $40				;VBL
 jp _vblank				;DMA routine

.pad $48				;disable LCDC status int
 reti

.pad $50				;disable timer overflow
 reti

.pad $58				;disable serial transfer int
 reti

.pad $60				;disable keypad oscillation int
 reti

_cpagerst:
 push af
 push hl
 ld hl,sp+6				; point to address on stack
 ldi a,(hl)				; load lsb of address
 ld h,(hl)				; load msb of address
 ld l,a					; complete pointer  
 inc hl					; past RST instr
 ldi a,(hl)				; get lsb of address
 ldh ($81),a			; load into high ram
 ldi a,(hl)				; get msb of address
 ldh ($82),a			; load into high ram
 ld a,(hl)				; get page to switch to
 ld ($2000),a			; switch rom pages
 pop hl
 pop af
 ret

_cpagejp_:
 add sp,2				; clear return address
 push af
 ld a,$c3				; opcode for JP
 ldh ($80),a			; load into high ram
 pop af
 jp $ff80				; jump to instruction in high ram

_cpagecall_:
 push af
 ld a,$cd				; opcode for CALL
 ldh ($80),a			; load into high ram
 pop af
 jp $ff80				; jump to instruction in high ram

_cpageldhl_:
 push af
 ldh a,($81)			; load from high ram
 ld l,a					; into lsb
 ldh a,($82)			; load from high ram
 ld h,a					; into msb
 pop af
 ret					; back to caller

.pad $100
 nop
 jp _main

;scrolling nintendo graphic
.db $CE,$ED,$66,$66,$CC,$0D,$00,$0B,$03,$73,$00,$83,$00,$0C,$00,$0D
.db $00,$08,$11,$1F,$88,$89,$00,$0E,$DC,$CC,$6E,$E6,$DD,$DD,$D9,$99
.db $BB,$BB,$67,$63,$6E,$0E,$EC,$CC,$DD,$DC,$99,$9F,$BB,$B9,$33,$3E

.db "CART NAME"		;16 bytes
.pad $143
.db $80				;gameboy color
.dw 0				;licensee code
.db 0				;gameboy
.db $1b				;ROM+MBC5+RAM+BATT
.db $00				;ROM Size (set by linker)
.db $02				;RAM Size (8kb)
.db 1				;non-japanese
.db $33				;check $144/$145 for license code
.db $00				;mask rom version #
.db $ff				;complement check
.db $ff,$ff       	;checksum

.end
