//---------------------------------------------------------------------------
#pragma hdrstop
#include <string.h>
#include "Link.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

bool CLink::Link(const char *gb, char *fname[], int fnames)
{
  FILE *fp;
  int files;

  error = false;
  _objs = 0;
  _sym = new _LSymEnt[MAXOSYM];
  _symnum = 0;
  for(files = 0; files < fnames; files++)
  {
    if ((fp = fopen(fname[files], "rb")) == NULL)
    {
      Error("cannot open object file: %s", fname[files]);
      return false;
    }
    Message("Loading object file \"%s\"", fname[files]);
    _obj[_objs] = new _Obj;
    strcpy(_obj[_objs]->fname, fname[files]);
    ReadObjectFile(fp);
    fclose(fp);
    _objs++;
  }

  if ((fp = fopen(gb, "wb")) == NULL)
  {
    Error("cannot open output file: %s", gb);
    return false;
  }
  Message("Writing ROM image %s", gb);
  Write(fp);
  fclose(fp);
//  Dump();

  delete []_sym;
  for(files = 0; files < _objs; files++)
    delete _obj[files];

  return true;
}

void CLink::Error(const char *m, const char *s)
{
  char str[256];

  sprintf(str, m, s);
  printf("Linker: %s\n", str);
  error = true;
}

void CLink::Message(const char *m, const char *s)
{
  char str[256];

  sprintf(str, m, s);
  printf("%s\n", str);
}

void CLink::ReadObjectFile(FILE *fp)
{
  int i, x;

  _obj[_objs]->outputsize = fgetw(fp);
  fread(_obj[_objs]->output, _obj[_objs]->outputsize, 1, fp);
  _obj[_objs]->loc.addr = fgetw(fp);
  for(i = 0; i < _obj[_objs]->loc.addr; i++)
    _obj[_objs]->loc.address[i] = fgetw(fp);
  _obj[_objs]->extsize = fgetw(fp);
  for(x = 0; x < _obj[_objs]->extsize; x++)
  {
    fread(_obj[_objs]->ext[x].name, fgetc(fp) + 1, 1, fp);
    _obj[_objs]->ext[x].addr = fgetw(fp);
    for(i = 0; i < _obj[_objs]->ext[x].addr; i++)
      _obj[_objs]->ext[x].address[i] = fgetw(fp);
  }
  _obj[_objs]->lsymsize = fgetw(fp);
  for(i = 0; i < _obj[_objs]->lsymsize; i++)
  {
    fread(_obj[_objs]->lsym[i].name, fgetc(fp) + 1, 1, fp);
    _obj[_objs]->lsym[i].addr = fgetw(fp);
  }
  _obj[_objs]->linenums = fgetw(fp);
  for(i = 0; i < _obj[_objs]->linenums; i++)
  {
    _obj[_objs]->linenum[i].line = fgetw(fp);
    _obj[_objs]->linenum[i].addr = fgetw(fp);
  }
}

void CLink::Dump()
{
  int i, x, n;

  for(n = 0; n < _objs; n++)
  {
    printf("File name: %s\n", _obj[n]->fname);
    printf("Output size: %i\n", _obj[n]->outputsize);
    printf("Page: %02x\n", _obj[n]->page);
    printf("Starting address: %04x\n", _obj[n]->addr);
    printf("Local relocation table:\n");
    for(i = 0; i < _obj[n]->loc.addr; i++)
      printf("  Addr: %04x\n", _obj[n]->loc.address[i]);
    printf("External relocation table:\n");
      for(i = 0; i < _obj[n]->extsize; i++)
    {
      printf("  Var: %s\n", _obj[n]->ext[i].name);
      for(x = 0; x < _obj[n]->ext[i].addr; x++)
        printf("    Addr: %04x\n", _obj[n]->ext[i].address[x]);
    }
    printf("Local symbols:\n");
    for(i = 0; i < _obj[n]->lsymsize; i++)
      printf("  Addr: %04x    Name: %s\n",
              _obj[n]->lsym[i].addr, _obj[n]->lsym[i].name);
/*
    printf("Line numbers:\n");
      for(i = 0; i < _obj[n]->linenums; i++)
        printf("  Line: %5i    Page: %02x    Addr: %04x\n",
                _obj[n]->linenum[i].line, _obj[n]->linenum[i].page,
                _obj[n]->linenum[i].addr);
*/
    printf("\n");
  }
  printf("Public symbols:\n");
  for(i = 0; i < _symnum; i++)
    printf("  Page: %02x    Addr: %04x    Name: %s\n",
            _sym[i].page, _sym[i].addr, _sym[i].name);
  printf("\n");
}

void CLink::Write(FILE *fp)
{
  const int BANKS[] = {2, 4, 8, 16, 32, 64, 128};
  char *rom[MAXBANKS],
       *c,
       line[128];
  int  page,
       addr,
       i, x, n,
       p, a;

  addr = 0;
  page = 0;
  for(i = 0; i < _objs; i++)
  {
    if ((_obj[i]->outputsize + addr > BANKSIZE) || (i == 0))
    {
      if (i > 0)
      {
        memset(&rom[page][addr], 0xFF, BANKSIZE - addr);
        addr = 0;
        page++;
        if (page == MAXBANKS)
        {
          sprintf(line, "output file exceeds %i banks!", MAXBANKS);
          Error(line, "");
        }
      }
      rom[page] = new char[BANKSIZE];
    }
    addr += AddPage(i, rom[page], addr, page);
  }
  memset(&rom[page][addr], 0xFF, BANKSIZE - addr);

  for(i = 0; i < _objs; i++)
    for(x = 0; x < _obj[i]->extsize; x++)
    {
      n = GetSym(_obj[i]->ext[x].name);
      if (n == -1)
      {
        Error("external symbol \"%s\" not found!\n",
              _obj[i]->ext[x].name);
        return;
      }
      a = _sym[n].addr;
      p = _sym[n].page;
      if ((p != 0) && (_obj[i]->page != 0))
        Error("cannot relocate load across pages in same address space!", "");
      for(n = 0; n < _obj[i]->ext[x].addr; n++)
      {
        c = &rom[_obj[i]->page][_obj[i]->addr + _obj[i]->ext[x].address[n]];
        if ((p == _obj[i]->page) || (p == 0))
          AddWord(c, a);
        else
        {
          c--;
          switch (*c)
          {
            case 0x21:
              *c++ = 0xFF;
              break;
            case 0xCD:
              *c++ = 0xF7;
              break;
            case 0xC3:
              *c++ = 0xEF;
              break;
          }
          AddWord(c, a);
          c += 2;
          *c = (char)p;
        }
      }
    }

  if (error)
    return;

  if (page < 2)
    x = 0;
  else if (page < 4)
    x = 1;
  else if (page < 8)
    x = 2;
  else if (page < 16)
    x = 3;
  else if (page < 32)
    x = 4;
  else if (page < 64)
    x = 5;
  else
    x = 6;
  rom[0][0x148] = (char)x;

  n = 0;
  for(i = 0x134; i < 0x14D; i++)
    n += rom[0][i];
  n = 0xE7 - (unsigned char)n;
  rom[0][0x14D] = (char)n;

  n = 0;
  for(i = 0; i < 0x14E; i++)
    n += (unsigned char)rom[0][i];
  for(i = 0x150; i < BANKSIZE; i++)
    n += (unsigned char)rom[0][i];
  for(p = 1; p <= page; p++)
    for(i = 0; i < BANKSIZE; i++)
      n += (unsigned char)rom[p][i];
  n += (BANKS[x] - page - 1) * (0xFF * BANKSIZE);
  rom[0][0x14E] = (char)(n >> 8);
  rom[0][0x14F] = (char)n;

  for(i = 0; i <= page; i++)
    fwrite(rom[i], BANKSIZE, 1, fp);
  memset(rom[0], 0xFF, BANKSIZE);
  for(i = page + 1; i < BANKS[x]; i++)
    fwrite(rom[0], BANKSIZE, 1, fp);

  for(i = 0; i <= page; i++)
    delete []rom[i];
}

int CLink::AddPage(int n, char *o, int s, int p)
{
  int i;

  memcpy(&o[s], _obj[n]->output, _obj[n]->outputsize);
  _obj[n]->addr = s;
  _obj[n]->page = p;
  for(i = 0; i < _obj[n]->loc.addr; i++)
    AddWord(&o[s + _obj[n]->loc.address[i]], s);
  for(i = 0; i < _obj[n]->lsymsize; i++)
    AddSym(_obj[n]->lsym[i].name, _obj[n]->lsym[i].addr + s, p);
  for(i = 0; i < _obj[n]->linenums; i++)
  {
    _obj[n]->linenum[i].addr += s;
    _obj[n]->linenum[i].page = p;
  }

  return _obj[n]->outputsize;     
}

void CLink::AddSym(const char *c, int v, int p)
{
  if (GetSym(c) != -1)
  {
    Error("duplicate public symbol \"%s\"", c);
    return;
  }

  strcpy(_sym[_symnum].name, c);
  _sym[_symnum].addr = v;
  _sym[_symnum].page = p;
  _symnum++;
}

int CLink::GetSym(const char *c)
{
  int i;

  for(i = 0; i < _symnum; i++)
    if (!strcmp(_sym[i].name, c))
      return i;

  return -1;
}

int CLink::fgetw(FILE *fp)
{
  unsigned char i[2];

  fread(i, 2, 1, fp);

  return i[0] + (i[1] << 8);
}

void CLink::AddWord(char *b, int a)
{
  *((short *)b) += (short)a; 
}
